from django.shortcuts import render, get_object_or_404, redirect
from django.contrib.auth.decorators import login_required
from django.utils import timezone
from django.db.models.fields import DateTimeField
from django.db.models import F, Count, Q, ExpressionWrapper, Case, When
from .models import ShiftRegistration, Room, Shift, Helper, Message, IncomingMessage
from django.views.generic import DetailView, ListView
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib import messages
from django.db import models
from django.core.paginator import Paginator
from .forms import BulkMessage, HelperShift


# Create your views here.
def index(request):
    pass


@login_required
def shift_overview(request):
    context = {}
    context["running_shifts"] = [
        shift
        for shift in Shift.objects.annotate(
            end_at=ExpressionWrapper(
                F("start_at") + F("duration"),
                output_field=DateTimeField(),
            )
        )
        .filter(start_at__lte=timezone.now(), end_at__gte=timezone.now(), deleted=False)
        .order_by("start_at")
    ]

    # only Postgres supports DISTINCT on specific columns, SQLite does not support aggregates on datetime fields
    context["next_shifts"] = filter(
        lambda x: x is not None,
        (
            Shift.objects.filter(room=room, start_at__gt=timezone.now(), deleted=False)
            .order_by("start_at")
            .first()
            for room in Room.objects.all()
        ),
    )

    return render(request, "shift_overview.html", context)


def add_helper_shift(self):
    pass


@login_required
def shift_detail(request, pk):
    shift = get_object_or_404(Shift, pk=pk)
    form = HelperShift()
    if request.method == "POST":
        form = HelperShift(request.POST)
        if form.is_valid():
            (reg, created) = ShiftRegistration.objects.get_or_create(
                helper=form.cleaned_data["helper"], shift=shift
            )
            if created:
                messages.add_message(
                    request,
                    messages.SUCCESS,
                    "Helfer erfolgreich zur Schicht hinzugefügt",
                )
            else:
                messages.add_message(
                    request,
                    messages.WARNING,
                    "Helfer ist bereits für diese Schicht angemeldet",
                )
            return redirect("team:shift", pk=shift.pk)
    # break potential cyclic imports
    from shiftregister.signage.forms import TeamBackupForm

    context = {
        "shift": shift,
        "add_helper_form": form,
        "backup_form": TeamBackupForm(),
    }
    return render(request, "shift_detail.html", context)


@login_required
def bulk_message(request):
    form = BulkMessage()

    if request.method == "POST":
        form = BulkMessage(request.POST)
        if form.is_valid():
            helpers = Helper.objects.filter(number_validated=True)
            if form.cleaned_data["checked_in_only"]:
                helpers = Helper.objects.annotate(
                    shift_count=Count(
                        Case(
                            When(
                                shiftregistration__state__in=[
                                    ShiftRegistration.RegState.CHECKED_IN,
                                ],
                                then=1,
                            ),
                            output_field=models.IntegerField(),
                        )
                    )
                ).filter(number_validated=True, shift_count__gte=1)

            try:
                Message.objects.bulk_create(
                    (
                        Message(
                            text=form.cleaned_data["message"].replace(
                                "$token",
                                f"https://helfen.kntkt.de{helper.logintoken_set.first().get_absolute_url()}",
                            ),
                            to=helper,
                        )
                        for helper in helpers
                    )
                )
                messages.add_message(
                    request, messages.SUCCESS, "Massen-Nachricht erfolgreich versendet"
                )
            except:
                messages.add_message(
                    request,
                    messages.ERROR,
                    "Fehler beim Versenden der Massen-Nachricht",
                )

    context = {
        "form": form,
    }
    return render(request, "bulk_message.html", context)


class HelperDetail(LoginRequiredMixin, DetailView):
    template_name = "helper_detail.html"
    model = Helper


class ShiftList(LoginRequiredMixin, ListView):
    template_name = "shift_list.html"
    model = Shift
    title = "Alle Schichten"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["title"] = self.title
        return context

    def get_ordering(self):
        return ("start_at", "room_id")


class FreeShiftList(ShiftList):
    title = "Freie Schichten"

    def get_queryset(self):
        help_wanted = Q(required_helpers__gt=F("reg_count")) | Q(
            required_helpers=0
        ) & Q(room__required_helpers__gt=F("reg_count"))
        return (
            Shift.with_reg_count()
            .annotate(
                end_at=ExpressionWrapper(
                    F("start_at") + F("duration"),
                    output_field=DateTimeField(),
                )
            )
            .filter(
                help_wanted,
                end_at__gte=timezone.now(),
                deleted=False,
            )
            .order_by("start_at", "room_id")
        )


@login_required
def checkin(request, pk):
    reg = get_object_or_404(ShiftRegistration, pk=pk)
    reg.state = reg.RegState.CHECKED_IN
    reg.save()
    return redirect("team:shift", pk=reg.shift.pk)


@login_required
def delete_shiftregistration(request, pk):
    reg = get_object_or_404(ShiftRegistration, pk=pk)
    spk = reg.shift.pk
    reg.delete()
    return redirect("team:shift", pk=spk)


@login_required
def incoming_messages(request):
    p = Paginator(IncomingMessage.objects.order_by("read", "-created_at"), 10)
    page = p.get_page(request.GET.get("page"))

    return render(
        request,
        "incoming_messages.html",
        {
            "num_unread": IncomingMessage.objects.filter(read=False).count(),
            "page": page,
            "page_range": p.get_elided_page_range(page.number),
        },
    )


@login_required
def conversation(request, sender):
    pass
